/* Include the controller definition */
#include "footbot_physarum.h"
/* Function definitions for XML parsing */
#include <argos3/core/utility/configuration/argos_configuration.h>
/* 2D vector definition */
#include <argos3/core/utility/math/vector2.h>
/* Function definitions for logging */
#include <argos3/core/utility/logging/argos_log.h>
#include <math.h>
#include <string>
/****************************************/
/****************************************/

CFootBotPhysarum::CFootBotPhysarum() :
   m_pcWheels(NULL),
   m_pcLEDs(NULL),
   m_pcProximity(NULL),
   m_pcDSA(NULL),
   m_pcDSS(NULL),
   m_pcRABS(NULL),
   m_pcRABA(NULL),
   m_cAlpha(10.0f),
   m_fDelta(0.5f),
   m_fWheelVelocity(10.0f),     // cm/s
   m_cGoStraightAngleRange(-ToRadians(m_cAlpha),
                           ToRadians(m_cAlpha)) {}

/****************************************/
/****************************************/

void CFootBotPhysarum::Init(TConfigurationNode& t_node) {
   /*
    * Get sensor/actuator handles
    *
    * The passed string (ex. "differential_steering") corresponds to the
    * XML tag of the device whose handle we want to have. For a list of
    * allowed values, type at the command prompt:
    *
    * $ argos3 -q actuators
    *
    * to have a list of all the possible actuators, or
    *
    * $ argos3 -q sensors
    *
    * to have a list of all the possible sensors.
    *
    * NOTE: ARGoS creates and initializes actuators and sensors
    * internally, on the basis of the lists provided the configuration
    * file at the <controllers><footbot_diffusion><actuators> and
    * <controllers><footbot_diffusion><sensors> sections. If you forgot to
    * list a device in the XML and then you request it here, an error
    * occurs.
    */
   m_pcWheels          = GetActuator<CCI_DifferentialSteeringActuator  >("differential_steering"   );
   m_pcProximity       = GetSensor  <CCI_FootBotProximitySensor        >("footbot_proximity"       );
   m_pcLEDs            = GetActuator<CCI_LEDsActuator                  >("leds"                    );

   m_pcDSA             = GetActuator<CCI_FootBotDistanceScannerActuator>("footbot_distance_scanner");
   m_pcDSS             = GetSensor  <CCI_FootBotDistanceScannerSensor  >("footbot_distance_scanner");

   m_pcRABS            = GetSensor  <CCI_RangeAndBearingSensor         >("range_and_bearing"       );
   m_pcRABA            = GetActuator<CCI_RangeAndBearingActuator       >("range_and_bearing"       );
   /*
    * Parse the configuration file
    *
    * The user defines this part. Here, the algorithm accepts three
    * parameters and it's nice to put them in the config file so we don't
    * have to recompile if we want to try other settings.
    */
   GetNodeAttributeOrDefault(t_node, "alpha", m_cAlpha, m_cAlpha);
   m_cGoStraightAngleRange.Set(-ToRadians(m_cAlpha), ToRadians(m_cAlpha));
   GetNodeAttributeOrDefault(t_node, "delta", m_fDelta, m_fDelta);
   GetNodeAttributeOrDefault(t_node, "velocity", m_fWheelVelocity, m_fWheelVelocity);
   m_pcDSA->Enable();
   m_pcDSA->SetRPM(300); // rounds per minutes

   Init_RABA_Data();

   /* Perform further initialization */
   Reset();

}

/****************************************/
/****************************************/

void CFootBotPhysarum::ControlStep() {
// if(m_id == 1 || m_id == 13)
// {
//    std::cout << "id:" << m_id << "type:" << m_robot_type << "\n";
// }
   switch(m_robot_type) {
      case STATE_EXPLORER:
         Explorer_Motion();
         break;
      case STATE_NODE:
         Node_Motion();
         break;
      case STATE_LOST_CHAIN:
         Lost_Chain_Motion();
         break;
      case STATE_TEMPORARY:
         Temporary_Motion();
         break;
      default:
         LOGERR << "[BUG] Unknown robot state: " << m_robot_type << std::endl;
   }
}


/****************************************************/
/*       Footbot get data of distance scanner       */
/****************************************************/
void CFootBotPhysarum::Get_Distance_Scanner()
{
   // const CCI_FootBotDistanceScannerSensor::TReadingsMap& tReads = m_ds->GetShortReadingsMap();
   /* Get readings from distance scanner sensor */
   const CCI_FootBotDistanceScannerSensor::TReadingsMap& tDisranceReads = m_pcDSS->GetReadingsMap();
   CVector2 cAccumulator;
   for(CCI_FootBotDistanceScannerSensor::TReadingsMap::const_iterator it = tDisranceReads.begin(); it != tDisranceReads.end(); ++it)
   {	
      // printf("%f\t%f\t\n",it->second, it->first.GetValue());   
      // const_iterator
      std::cout << "distance  :  " << it->second << "angle" << it->first.GetValue()  << "\n";
      std::cout << "distance  :  " << "\n";
      cAccumulator += CVector2(it->second, it->first);
   }
}

/****************************************/
/*       Footbot Explorer motion        */
/****************************************/
void CFootBotPhysarum::Explorer_Motion(){
   if (m_robot_type != STATE_EXPLORER)
   {
      m_robot_type = STATE_EXPLORER;
      m_pcRABA->SetData(0, STATE_EXPLORER);
   }
   m_pcLEDs->SetAllColors(CColor::RED);
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   if(! tMsgs.empty()) {
      /* Used to calculate the sum gradient of each node of robot chain's contribution */
      Real sum_gradient = 0;
      /* Used to calculate the number of each neighbor node of robot chain */
      Real sum_size     = 0;
      /* Used to calculate the field strength */
      Real field_strength;
      /*Used to storage the index of the node index if there is only one node*/
      size_t index = 255;

      UInt8 flag_close_end = 0;
      /* Caluculate the number of nodes in robot chain and the average value of gradient*/
      for(size_t i = 0; i < tMsgs.size(); ++i) {
         if(tMsgs[i].Data[0] == STATE_NODE ||  tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
         {
            sum_gradient += tMsgs[i].Data[1];
            sum_size++;
            index = i;

            if(tMsgs[i].Data[6] == 255)
            {
               flag_close_end = 1;
            }
         }
      }
      if(sum_size == 1)
      {
         Extension_mechanism(sum_size, index);
      }
      else if(sum_size == 0)
      {
         /*transform to STATE_LOST_CHAIN*/
         Lost_Chain_Motion();
      }
      else{
         Real aver_gradient = sum_gradient/sum_size;
         /*Calculate the guiding direction of potential field*/
         CVector2 potential_direction = GeneralizedPotential(aver_gradient,sum_size);
         // std::cout << "p_dir:" << potential_direction << "\n";
         /*Calculate the advoid obstacles direction*/
         CVector2 advoid_direction    = Generalized_Dynamic_Advoid_Obstacles();
         // CVector2 advoid_direction = repulsion_vector();

         Real B_factor                = Get_Balance_Factor();
         /*Sum all directions*/
         CVector2 sum_direction       = (1-B_factor)*potential_direction + B_factor*advoid_direction;
         // if(m_id == 3)  std::cout <<  "a_dir:" << advoid_direction << "s_dir:" << sum_direction << "\n";
         /*Moving according to the direction*/
         // if sum_direction.GetY()
         // Real temp = sum_direction.GetX()*sum_direction.GetX()+sum_direction.GetY()*sum_direction.GetY();
         // if ( temp > 0.001)
         // {
         //    Vector_to_Wheel_Velocity_Noscale(sum_direction);
         // }
         // else
         // {
         //    m_pcWheels->SetLinearVelocity(m_fWheelVelocity,m_fWheelVelocity);
         // }
         Vector_to_Wheel_Velocity_Noscale(sum_direction);
      }
   }
   else {
      LOG << "none";
      /*transform to STATE_LOST_CHAIN*/
      Lost_Chain_Motion();
   }
}

/*The condition for running this function is that there is only one robot chain node around Explorer*/
void CFootBotPhysarum::Extension_mechanism(UInt8 sum_size,UInt8 index)
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   Real B_factor = Get_Balance_Factor();
   
   // if(sum_size == 1)
   // {
      if((tMsgs[index].Range > u_length_rc && B_factor == 0) && tMsgs[index].Data[5] != 0)
      {
         // if(m_id == 5 || m_id == 1) std::cout << "id:" << m_id << "last_n:" << tMsgs[index].Data[2] << "\n";
         /*transform to STATE_NODE*/
         P_node = tMsgs[index].Data[2];
         Temporary_Motion();
      }
      else
      {
         /*Calculate the advoid obstacles direction*/
         CVector2 advoid_direction = Generalized_Dynamic_Advoid_Obstacles();
         // CVector2 advoid_direction = repulsion_vector();
         /*Generate a random vector*/
         CVector2 random_direction = CVector2((Real)(rand()%10)/10,(Real)(rand()%10)/10);
         // std::cout << (Real)(rand()%10)/10 << "\n";
         CVector2 sum_direction = advoid_direction+random_direction;
         /*Moving according to the direction*/
         // Real temp = sum_direction.GetX()*sum_direction.GetX()+sum_direction.GetY()*sum_direction.GetY();
         // if ( temp > 0.001)
         // {
         //    Vector_to_Wheel_Velocity_Noscale(sum_direction);
         // }
         // else
         // {
         //    m_pcWheels->SetLinearVelocity(m_fWheelVelocity,m_fWheelVelocity);
         // }
         Vector_to_Wheel_Velocity_Noscale(advoid_direction);
      }
}

/****************************************/
/*         Footbot Node Motion          */
/****************************************/
void CFootBotPhysarum::Node_Motion(){
   if (m_robot_type != STATE_NODE)
   {
      m_robot_type = STATE_NODE;
      m_pcRABA->SetData(0, m_robot_type);
   }
   // if(m_id == 1 || m_id == 5) std::cout << "id:" << m_id << "c_id:" << C_node <<  "g:" << m_node_gradient << "P_node:" << P_node << "r_id:" << m_root_id << "e_f:" << e_f << "\n";
   if(e_f == 1)
   {
      m_pcLEDs->SetAllColors(CColor::BLUE);
   }
   else
   {
      m_pcLEDs->SetAllColors(CColor::WHITE);
   }
   
   Check_Chain();
   /*Update the gradient of node*/
   Update_gradient();

   Update_sum_c_node();
   Update_Root_ID();
   Caculate_Envir_Factor();
   Spread_e_f();
}

/*Determine whether the current node is the end of the robot chain and then caculate the environmental assessment factors*/
void CFootBotPhysarum::Update_sum_c_node()
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   UInt8 sum_c_node = 0;
   UInt8 temp_c_node = 255;
   for(size_t i = 0; i < tMsgs.size();i++)
   {
      if(tMsgs[i].Data[0] == STATE_NODE || tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
      {
         // Get ID of the child node and caculate the number of child node
         if(tMsgs[i].Data[1] == m_node_gradient + 1)
         {
            sum_c_node++;
            temp_c_node = tMsgs[i].Data[2];
         }
      }
   }
   if (sum_c_node != num_C_node)
   {
      num_C_node = sum_c_node;
      m_pcRABA->SetData(3, num_C_node);
   }
   if(C_node != temp_c_node)
   {
      C_node = temp_c_node;
      m_pcRABA->SetData(6,C_node);
   }
}

void CFootBotPhysarum::Update_Root_ID()
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   for(size_t i = 0; i < tMsgs.size();i++)
   {
      if(tMsgs[i].Data[0] == STATE_NODE || tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
      {
         // Caculate the root_id
         if (tMsgs[i].Data[2] == P_node)
         {
            m_root_id = tMsgs[i].Data[4];
            if(tMsgs[i].Data[3] > 1)
            {
               m_root_id = m_id;       // update the root id
            }
         }
      }
   }
   // if (m_id == 14) std::cout << "id:" << m_id << "root_id:" << m_root_id << "\n";
   m_pcRABA->SetData(4,m_root_id);
}

/*Determine whether the current node is the end of the robot chain and then caculate the environmental assessment factors*/
void CFootBotPhysarum::Caculate_Envir_Factor()
{
   UInt8 flag_exist_explorer = 0;
   if (C_node == 255)               // The end of the robot chain
   {
      /* Get RAB messages from nearby foot-bots */
      const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
      for(size_t i = 0; i < tMsgs.size();i++)
      {
         if(tMsgs[i].Data[0] == STATE_EXPLORER)
         {
            flag_exist_explorer = 1;      // exist explorer
            break;
         }
      }
   }
   if(flag_exist_explorer == 1)
   {
      counter_envir += 1;
   }
   if(counter_envir > 1000)
   {
      e_f = 0;
      m_pcRABA->SetData(5,e_f);
      flag_exist_explorer = 0;
      counter_envir = 0;
   }
}

void CFootBotPhysarum::Spread_e_f()
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   if(C_node != 255)
   {
      for(size_t i=0;i<tMsgs.size();i++)
      {
         if(tMsgs[i].Data[4] == m_root_id && (tMsgs[i].Data[0] == STATE_NODE || tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY))
         {
            if(tMsgs[i].Data[2] ==  C_node)
            { 
               e_f = tMsgs[i].Data[5];
            }
         }
      }
      m_pcRABA->SetData(5,e_f);
   }
}

/****************************************/
/*    Footbot Temporary Node Motion     */
/****************************************/
void CFootBotPhysarum::Temporary_Motion()
{
   if (m_robot_type != STATE_TEMPORARY)
   {
      m_robot_type = STATE_TEMPORARY;
      m_pcRABA->SetData(0, m_robot_type);
   }

   m_pcLEDs->SetAllColors(CColor::YELLOW);

   m_pcWheels->SetLinearVelocity(0,0); //Stop

   counter += 1;        // Timer plus 1

   UInt8 num_node = 0;  // Init the number of node nearby

   Real dist_to_closest_node = 255;

   UInt8 flag_out = 1;
   UInt8 flag_exist_p_node = 0;

   // if(m_node_gradient = 255)
   // {
      Update_gradient();
   // }

   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();

   for(size_t i = 0; i < tMsgs.size(); i++)
   {
      if(tMsgs[i].Data[0] == 1)
      {
         // There is a target
         flag_target = 1;
      }
      else if(tMsgs[i].Data[0] == STATE_NODE || tMsgs[i].Data[0] == STATE_TEMPORARY || tMsgs[i].Data[0] == 0)
      {
         if(tMsgs[i].Data[5] == 0)
         {
            flag_out = 0;
            break;
         }
         if(tMsgs[i].Data[2] == P_node)
         {
            flag_exist_p_node = 1;
         }   
         if(tMsgs[i].Range < dist_to_closest_node && (tMsgs[i].Data[1] == m_node_gradient || tMsgs[i].Data[1] == m_node_gradient - 1))
         {
            dist_to_closest_node = tMsgs[i].Range;
         }
      }
   }

   if(flag_out == 0)
   {
      counter = 0;
      Node_to_Other_State();
      Explorer_Motion();
   }
   // if (m_id == 2) std::cout << "p_node:" << P_node << "exit_p:" << flag_exist_p_node << "\n";
   // -- If 10 seconds have expired, make a decision
   if(counter > 100)
   {
      counter = 0;
      if (flag_target == 1)
      {
         // Node_to_Other_State();
         m_node_gradient = 255;
         Node_Motion();
      }
      else if(flag_exist_p_node == 0)
      {
         Node_to_Other_State();
         Explorer_Motion();
      }
      else if(flag_target == 0 && dist_to_closest_node > 200)
      {

         // Node_to_Other_State();
         m_node_gradient = 255;
         Node_Motion();
      }
      else
      {
         Node_to_Other_State();
         Explorer_Motion();
      }
   }
}
/*
 * This function is a generalization of the potential
 */
CVector2 CFootBotPhysarum::GeneralizedPotential(Real aver_gradient, Real sum_size)
{
   Real field_strength;
   CVector2 potential_direction = {0,0};
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   for(size_t i = 0; i < tMsgs.size(); ++i) 
   {
      if(tMsgs[i].Data[0] == STATE_NODE ||  tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
      {
         field_strength = (tMsgs[i].Data[1] - aver_gradient)*log(tMsgs[i].Range)/tMsgs[i].Data[5];      // Caculate the potential field strength
         if(tMsgs[i].Data[5] == 1)
         {
            field_strength = (tMsgs[i].Data[1] - aver_gradient)*log(tMsgs[i].Range)/tMsgs[i].Data[5];
         }
         else
         {
            field_strength = -(tMsgs[i].Data[1] - aver_gradient)*log(tMsgs[i].Range);
         }
         potential_direction += CVector2(field_strength,tMsgs[i].HorizontalBearing);
         // if (field_strength)
         // std::cout << "field_strength:" << field_strength << "\n";
      }
   }
   potential_direction /=sum_size;
   return potential_direction;
}

/*Caculate the balance factor
*/
Real CFootBotPhysarum::Get_Balance_Factor()
{
   /* Get readings from proximity sensor */
   const CCI_FootBotProximitySensor::TReadings& tProxReads = m_pcProximity->GetReadings();

   Real max_Prox_Value = 0;

   for(size_t i=0; i< tProxReads.size();i++)
   {
      if(tProxReads[i].Value > max_Prox_Value)
      {
         max_Prox_Value = tProxReads[i].Value;
      }
   }
   return max_Prox_Value;
}

/*
 * This function is a generalization of the advoid obstacles based on Guassian
 */
 CVector2 CFootBotPhysarum::Generalized_Dynamic_Advoid_Obstacles()
 {
   /* Get readings from proximity sensor */
   const CCI_FootBotProximitySensor::TReadings& tProxReads = m_pcProximity->GetReadings();
   /*Store the sensing value of proximity sensor*/
   Real fitness_prox[tProxReads.size()]={0};  // Init the value of array = 0

   // std::memset(fitness_prox,0,sizeof(fitness_prox));

   UInt8 flag_obstacles = 0;     // 0 means no obstacle detected; 1 indicates an obstacle has been detected

   CVector2 advoid_direction;

   for(size_t i=0;i < tProxReads.size();++i)
   {
      /*Calculate the effect of other proximity sensors*/
      if(!m_cGoStraightAngleRange.WithinMinBoundIncludedMaxBoundIncluded(tProxReads[i].Angle) && tProxReads[i].Value > m_fDelta )
      {
         flag_obstacles = 1;
         for(size_t j=0; j < tProxReads.size();j++)
         {
            // calculate the effect of other proximity sensors based on the Gaussian function
            Real g_result = Generalized_Guassian(tProxReads[j].Angle.GetValue(),tProxReads[i].Angle.GetValue(),1);
            fitness_prox[j] += tProxReads[i].Value * g_result;
            // std::cout << i << "effort to" << j << "is:" << g_result << "\n";
         }
      }
   }

   /* find the minimum value of guassion function and the index of proximity*/
   Real min_g_result = 255;
   size_t min_g_result_index = 0;
   if(flag_obstacles != 0)
   {
      for(size_t i = 0; i < tProxReads.size(); i++)
      {
         if (fitness_prox[i] < min_g_result)
         {
            min_g_result = fitness_prox[i];
            min_g_result_index = i;
         }
         // std::cout<< "pro:" << i << "a_s:" << fitness_prox[i] << "prox_value:" << tProxReads[i].Value << "\n";
      }
      advoid_direction = CVector2(1,tProxReads[min_g_result_index].Angle);
      // std::cout << "min_index:" << min_g_result_index << "advoid_direction:" << advoid_direction << "\n";
   }
   else
   {
      advoid_direction = {0,0};
   }
   // std::cout << "a_strength:" << fitness_prox[min_g_result_index] << "\n";
   // std::cout << "direction:" << advoid_direction <<"\n";
   return advoid_direction;
 }

CVector2 CFootBotPhysarum::repulsion_vector()
{
   /* Get readings from proximity sensor */
   const CCI_FootBotProximitySensor::TReadings& tProxReads = m_pcProximity->GetReadings();

   /* Sum them together */
   /* length*cos(angle),length*sin(angle)*/
   CVector2 cAccumulator;
   for(size_t i = 0; i < tProxReads.size(); ++i) {
      cAccumulator += CVector2(tProxReads[i].Value, tProxReads[i].Angle);
   }
   cAccumulator = -cAccumulator/tProxReads.size();     // tProxReads.size() = 24
   /* If the angle of the vector is small enough and the closest obstacle
    * is far enough, continue going straight, otherwise curve a little
    */
   //  if(cAccumulator.Get*repulsion.x+repulsion.y*repulsion.y > 0.001)
   CRadians cAngle = cAccumulator.Angle();
   if(m_cGoStraightAngleRange.WithinMinBoundIncludedMaxBoundIncluded(cAngle) && cAccumulator.Length() < m_fDelta ) {
      cAccumulator = {0,0};
   }
   // std::cout << "vector:" << cAccumulator << "\n";
   return cAccumulator;
}

/*
 * This function is a generalization of the Guassian function
 * x -- the input value, m -- the mean value,   s --variance
 */
 Real CFootBotPhysarum::Generalized_Guassian(Real x, Real m, Real s)
 {
    static const Real inv_sqrt_2pi = 0.3989422804014327;
    Real a = (x - m) / s;

    return inv_sqrt_2pi / s * std::exp(-0.5f * a * a);
}

/* Transforms the given direction vector into wheel actuation
  This version does not consider the length of the direction vector*/
void CFootBotPhysarum::Vector_to_Wheel_Velocity_Noscale(CVector2 direction)
{
    // Get the direction angle in (-pi,pi)
   // Real angle = direction.Angle();
   Real angle = atan2(direction.GetY(),direction.GetX());
    // Declare speeds
   Real lvel, rvel;
   if (angle > 0)
   {
      lvel = m_fWheelVelocity * (1 - 2 * angle  / M_PI);
      rvel = m_fWheelVelocity;
   }
   else
   {
      lvel = m_fWheelVelocity;
      rvel = m_fWheelVelocity * (1 + 2 * angle  / M_PI);
   }
    // Set wheel speeds
    m_pcWheels->SetLinearVelocity(lvel, rvel);
   //  m_pcWheels->SetLinearVelocity(m_fWheelVelocity
}

/*Update the Gradient of node*/
void CFootBotPhysarum::Update_gradient()
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   UInt8 min_gradient = 255;
   UInt8 min_g_id = 255;
   if(! tMsgs.empty()) {
      for(size_t i = 0; i < tMsgs.size(); i++)
      {
         if(tMsgs[i].Data[0] == STATE_NODE ||  tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
         {
            if(tMsgs[i].Data[1] < min_gradient)
            {
               min_gradient = tMsgs[i].Data[1];
               min_g_id = tMsgs[i].Data[2];
            }
         }
      }
      
      if (m_node_gradient != min_gradient + 1 && min_gradient != 255)
      {
         m_node_gradient = min_gradient + 1;
         m_pcRABA->SetData(1, m_node_gradient);
         if(m_robot_type != STATE_TEMPORARY)         P_node = min_g_id;
         // std::cout << "gadient:" << m_node_gradient << "\n";
      }
   }
}

/*Determine if the robot chain is missing for the Robot_Node*/
void CFootBotPhysarum::Check_Chain()
{
   counter += 1;
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   
   if(! tMsgs.empty())
   {
      for(size_t i=0; i < tMsgs.size();i++)
      {
         if(tMsgs[i].Data[0] == STATE_NODE || tMsgs[i].Data[0] == STATE_TEMPORARY || tMsgs[i].Data[0] == 0)
         {

            if(tMsgs[i].Data[2] == P_node && tMsgs[i].Range > 200)
            {
               flag_loss = 0;
            }
            if(C_node != 255)
            {
               if(tMsgs[i].Data[2] == C_node)
               {
                  flag_c_loss = 0;
               }
            }
         }
      }
   }
   // else
   // {
   //    tansform to lost chain
   //    Node_to_Other_State();
   //    std::cout << "Lost the any communication" << "\n";
   //    Lost_Chain_Motion();
   // }
   if(counter > 100)
   {
      counter = 0;
      // The parent node is not sensed once in 10 seconds
      if(flag_loss == 1)
      {
         Node_to_Other_State();
         std::cout << "ID:" << m_id << "????????" <<"\n";
         Lost_Chain_Motion();
      }
      // The child node is not sensed once in 10 seconds
      if(flag_c_loss == 1)
      {
         C_node = 255;
      }
      
      flag_loss = 1;
      flag_c_loss = 1;
   }
}

void CFootBotPhysarum::Node_to_Other_State()
{
   m_node_gradient = 255;
   num_C_node = 0;
   m_root_id = 255;
   P_node = 255;
   C_node = 255;
   e_f = 1;

   m_pcRABA->SetData(1, m_node_gradient);
   m_pcRABA->SetData(3, num_C_node);
   m_pcRABA->SetData(4,m_root_id);
   m_pcRABA->SetData(5,e_f);
}

/****************************************/
/*          Footbot Lost Chain          */
/****************************************/
void CFootBotPhysarum::Lost_Chain_Motion()
{
   if (m_robot_type != STATE_LOST_CHAIN)
   {
      m_robot_type = STATE_LOST_CHAIN;
      m_pcRABA->SetData(0, m_robot_type);
   }
   m_pcLEDs->SetAllColors(CColor::GREEN);
   // m_pcWheels->SetLinearVelocity(0,0);

   /*Calculate the advoid obstacles direction*/
   CVector2 advoid_direction = Generalized_Dynamic_Advoid_Obstacles();
   // CVector2 advoid_direction = repulsion_vector();
   /*Moving according to the direction*/
   Vector_to_Wheel_Velocity_Noscale(advoid_direction);

   // Real temp = advoid_direction.GetX()*advoid_direction.GetX()+advoid_direction.GetY()*advoid_direction.GetY();
   // if (temp > 0.001)
   // {
   //    Vector_to_Wheel_Velocity_Noscale(advoid_direction);
   // }
   // else
   // {
   //    m_pcWheels->SetLinearVelocity(m_fWheelVelocity,m_fWheelVelocity);
   // }
   Lost_to_Explorer();
}

/*The condition of */
void CFootBotPhysarum::Lost_to_Explorer()
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();

   UInt8 flag_chain = 0;

   for(size_t i = 0; i < tMsgs.size(); i++)
   {
      if(tMsgs[i].Data[0] == STATE_NODE || tMsgs[i].Data[0] == STATE_TEMPORARY || tMsgs[i].Data[0] == 0)
      {
         flag_chain = 1;
         break;
      }
   }
   if(flag_chain == 1)
   {
      Explorer_Motion();
   }
}

/****************************************/
/*       Footbot advoid obstacles       */
/****************************************/
void CFootBotPhysarum::Advoid_obstacles() {
   /* Get readings from proximity sensor */
   const CCI_FootBotProximitySensor::TReadings& tProxReads = m_pcProximity->GetReadings();
   
   /* Sum them together */
   /* length*cos(angle),length*sin(angle)*/
   CVector2 cAccumulator;
   for(size_t i = 0; i < tProxReads.size(); ++i) {
      cAccumulator += CVector2(tProxReads[i].Value, tProxReads[i].Angle);
   }
   cAccumulator /= tProxReads.size();     // tProxReads.size() = 24
   /* If the angle of the vector is small enough and the closest obstacle
    * is far enough, continue going straight, otherwise curve a little
    */
   CRadians cAngle = cAccumulator.Angle();
   if(m_cGoStraightAngleRange.WithinMinBoundIncludedMaxBoundIncluded(cAngle) &&
      cAccumulator.Length() < m_fDelta ) {
      /* Go straight */
      m_pcWheels->SetLinearVelocity(m_fWheelVelocity, m_fWheelVelocity);
   }
   else {
      /* Turn, depending on the sign of the angle */
      if(cAngle.GetValue() > 0.0f) {
         m_pcWheels->SetLinearVelocity(m_fWheelVelocity, 0.0f);
      }
      else {
         m_pcWheels->SetLinearVelocity(0.0f, m_fWheelVelocity);
      }
   }
}

void CFootBotPhysarum::Reset()
{
   Init_RABA_Data();
}

void CFootBotPhysarum::Init_RABA_Data()
{
   m_pcRABA->ClearData();
   /* Switch to state explorer */
   m_robot_type = STATE_EXPLORER;
   /* Tell robots around that this robot is explorer */
   m_pcRABA->SetData(0, m_robot_type);

   const std::string& str = CCI_Controller::GetId();
   // std::cout<<str.find("r")<<"\n";
   std::string str3 = str.substr(8,str.length());
   // std::cout << str <<" "<< str3 << "\n";

   m_id = atoi(str3.c_str());

   m_pcRABA->SetData(2,m_id);

   m_pcRABA->SetData(5,e_f);

   m_pcRABA->SetData(6,C_node);
}

/*Get Id from string*/
// void extractIntegerWords(std::string str) 
// { 
   
// }

/****************************************/
/****************************************/

/*
 * This statement notifies ARGoS of the existence of the controller.
 * It binds the class passed as first argument to the string passed as
 * second argument.
 * The string is then usable in the configuration file to refer to this
 * controller.
 * When ARGoS reads that string in the configuration file, it knows which
 * controller class to instantiate.
 * See also the configuration files for an example of how this is used.
 */
REGISTER_CONTROLLER(CFootBotPhysarum, "footbot_physarum_controller")
