/* Include the controller definition */
#include "footbot_physarum.h"
/* Function definitions for XML parsing */
#include <argos3/core/utility/configuration/argos_configuration.h>
/* 2D vector definition */
#include <argos3/core/utility/math/vector2.h>
/* Function definitions for logging */
#include <argos3/core/utility/logging/argos_log.h>
#include <math.h>
#include <string>
/****************************************/
/****************************************/

CFootBotPhysarum::CFootBotPhysarum() :
   m_pcWheels(NULL),
   m_pcLEDs(NULL),
   m_pcProximity(NULL),
   m_pcDSA(NULL),
   m_pcDSS(NULL),
   m_pcRABS(NULL),
   m_pcRABA(NULL),
   m_pcMGS(NULL),
   m_cAlpha(10.0f),
   m_fDelta(0.5f),
   m_fWheelVelocity(10.0f),     // cm/s
   m_cGoStraightAngleRange(-ToRadians(m_cAlpha),
                           ToRadians(m_cAlpha)) {}

/****************************************/
/****************************************/

void CFootBotPhysarum::Init(TConfigurationNode& t_node) {
   /*
    * Get sensor/actuator handles
    *
    * The passed string (ex. "differential_steering") corresponds to the
    * XML tag of the device whose handle we want to have. For a list of
    * allowed values, type at the command prompt:
    *
    * $ argos3 -q actuators
    *
    * to have a list of all the possible actuators, or
    *
    * $ argos3 -q sensors
    *
    * to have a list of all the possible sensors.
    *
    * NOTE: ARGoS creates and initializes actuators and sensors
    * internally, on the basis of the lists provided the configuration
    * file at the <controllers><footbot_diffusion><actuators> and
    * <controllers><footbot_diffusion><sensors> sections. If you forgot to
    * list a device in the XML and then you request it here, an error
    * occurs.
    */
   m_pcWheels          = GetActuator<CCI_DifferentialSteeringActuator  >("differential_steering"   );
   m_pcProximity       = GetSensor  <CCI_FootBotProximitySensor        >("footbot_proximity"       );
   m_pcLEDs            = GetActuator<CCI_LEDsActuator                  >("leds"                    );

   m_pcDSA             = GetActuator<CCI_FootBotDistanceScannerActuator>("footbot_distance_scanner");
   m_pcDSS             = GetSensor  <CCI_FootBotDistanceScannerSensor  >("footbot_distance_scanner");

   m_pcRABS            = GetSensor  <CCI_RangeAndBearingSensor         >("range_and_bearing"       );
   m_pcRABA            = GetActuator<CCI_RangeAndBearingActuator       >("range_and_bearing"       );

   m_pcMGS             = GetSensor  <CCI_FootBotMotorGroundSensor      >("footbot_motor_ground"    );
   /*
    * Parse the configuration file
    *
    * The user defines this part. Here, the algorithm accepts three
    * parameters and it's nice to put them in the config file so we don't
    * have to recompile if we want to try other settings.
    */
   GetNodeAttributeOrDefault(t_node, "alpha", m_cAlpha, m_cAlpha);
   m_cGoStraightAngleRange.Set(-ToRadians(m_cAlpha), ToRadians(m_cAlpha));
   GetNodeAttributeOrDefault(t_node, "delta", m_fDelta, m_fDelta);
   GetNodeAttributeOrDefault(t_node, "velocity", m_fWheelVelocity, m_fWheelVelocity);
   m_pcDSA->Enable();
   m_pcDSA->SetRPM(300); // rounds per minutes

   Init_RABA_Data();

   /* Perform further initialization */
   Reset();

}

/****************************************/
/****************************************/

void CFootBotPhysarum::ControlStep() {
// if(m_id == 1 || m_id == 13)
// {
//    std::cout << "id:" << m_id << "type:" << m_robot_type << "\n";
// }
// if(m_id == 19) std::cout << "type:" << m_robot_type << "\n";
   switch(m_robot_type) {
      case STATE_EXPLORER:
         Explorer_Motion();
         break;
      case STATE_NODE:
         Node_Motion();
         break;
      case STATE_LOST_CHAIN:
         Lost_Chain_Motion();
         break;
      case STATE_TEMPORARY:
         Temporary_Motion();
         break;
      case STATE_WANDER:
         wander_in_waiting_area();
         break;
      case STATE_EXIST:
         Exist_Motion();
         break;
      default:
         LOGERR << "[BUG] Unknown robot state: " << m_robot_type << std::endl;
   }
}


/****************************************************/
/*       Footbot get data of distance scanner       */
/****************************************************/
void CFootBotPhysarum::Get_Distance_Scanner()
{
   // const CCI_FootBotDistanceScannerSensor::TReadingsMap& tReads = m_ds->GetShortReadingsMap();
   /* Get readings from distance scanner sensor */
   const CCI_FootBotDistanceScannerSensor::TReadingsMap& tDisranceReads = m_pcDSS->GetReadingsMap();
   CVector2 cAccumulator;
   for(CCI_FootBotDistanceScannerSensor::TReadingsMap::const_iterator it = tDisranceReads.begin(); it != tDisranceReads.end(); ++it)
   {	
      // printf("%f\t%f\t\n",it->second, it->first.GetValue());   
      // const_iterator
      std::cout << "distance  :  " << it->second << "angle" << it->first.GetValue()  << "\n";
      std::cout << "distance  :  " << "\n";
      cAccumulator += CVector2(it->second, it->first);
   }
}

/****************************************/
/*       Footbot Explorer motion        */
/****************************************/
void CFootBotPhysarum::Explorer_Motion(){
   if (m_robot_type != STATE_EXPLORER)
   {
      m_robot_type = STATE_EXPLORER;
      m_pcRABA->SetData(0, STATE_EXPLORER);
   }
   if (!is_out_waiting_area())
   {
      wander_in_waiting_area();
      return;
   }
   // is_out_waiting_area();
   m_pcLEDs->SetAllColors(CColor::RED);
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   if(! tMsgs.empty()) {
      /* Used to calculate the sum gradient of each node of robot chain's contribution */
      Real sum_gradient = 0;
      /* Used to calculate the number of each neighbor node of robot chain */
      Real sum_size     = 0;
      /* Used to calculate the field strength */
      Real field_strength;
      /*Used to storage the index of the node index if there is only one node*/
      UInt8 index = 255;

      /* Caluculate the number of nodes in robot chain and the average value of gradient*/
      for(size_t i = 0; i < tMsgs.size(); ++i) {
         if(tMsgs[i].Data[0] == STATE_NODE ||  tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
         {
            sum_gradient += tMsgs[i].Data[1];
            sum_size++;
            index = i;
         }
      }
      if(sum_size == 1)
      {
         Extension_mechanism(sum_size, index);
      }
      else if(sum_size == 0)
      {
         /*transform to STATE_LOST_CHAIN*/
         Lost_Chain_Motion();
      }
      else{
         Real aver_gradient = sum_gradient/sum_size;
         /*Calculate the guiding direction of potential field*/
         CVector2 potential_direction = GeneralizedPotential(aver_gradient,sum_size);
         // if (m_id == 20) std::cout << "id:" << m_id<< "p_dir:" << potential_direction << "\n";
         /*Calculate the advoid obstacles direction*/
         // CVector2 advoid_direction    = Generalized_Dynamic_Advoid_Obstacles();
         CVector2 advoid_direction = repulsion_vector();

         Real B_factor                = Get_Balance_Factor();
         /*Sum all directions*/
         CVector2 sum_direction       = (1-B_factor)*potential_direction + B_factor*advoid_direction;
         /*Moving according to the direction*/

         Real temp = sum_direction.GetX()*sum_direction.GetX()+sum_direction.GetY()*sum_direction.GetY();
         if ( temp > 0.001)
         {
            Vector_to_Wheel_Velocity_Noscale(sum_direction);
         }
         else
         {
            m_pcWheels->SetLinearVelocity(m_fWheelVelocity,m_fWheelVelocity);
         }
      }
   }
   else {
      LOG << "none";
      /*transform to STATE_LOST_CHAIN*/
      Lost_Chain_Motion();
   }
}

/*The condition for running this function is that there is only one robot chain node around Explorer*/
void CFootBotPhysarum::Extension_mechanism(UInt8 sum_size,UInt8 index)
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   Real B_factor = Get_Balance_Factor();
   // && tMsgs[index].Data[5] != 0
   // && B_factor == 0 
   // tMsgs[index].Data[5] != 0
   // if(index != 255){
      // if(m_id == 9 || m_id == 5 || m_id == 2) std::cout << "id:" << m_id << "index:" << index << "P:" << tMsgs[index].Data[2] << "g:" << tMsgs[index].Data[1] + 1 << "dist:" << tMsgs[index].Range << "\n";
      if((tMsgs[index].Range > 0.7*RAB_RANGE && tMsgs[index].Range < 0.8*RAB_RANGE) && is_out_waiting_area())
      {
         /*transform to STATE_NODE*/
         P_node = tMsgs[index].Data[2];
         m_node_gradient = tMsgs[index].Data[1] + 1;
         m_pcRABA->SetData(1,m_node_gradient);
         Temporary_Motion();
         return;
      }
      else
      {
         /*Calculate the advoid obstacles direction*/
         // CVector2 advoid_direction = Generalized_Dynamic_Advoid_Obstacles();
         CVector2 advoid_direction = repulsion_vector();
         /*Generate a random vector*/

         // CVector2 random_direction = CVector2((Real)(rand()%10)/10,(Real)(rand()%10)/10);
         // CVector2 sum_direction = advoid_direction+random_direction;

         /*Moving according to the direction*/
         // Vector_to_Wheel_Velocity_Noscale(advoid_direction);

         Real temp = advoid_direction.GetX()*advoid_direction.GetX()+advoid_direction.GetY()*advoid_direction.GetY();
         if ( temp > 0.001)
         {
            Vector_to_Wheel_Velocity_Noscale(advoid_direction);
         }
         else
         {
            m_pcWheels->SetLinearVelocity(m_fWheelVelocity,m_fWheelVelocity);
         }
      }
   // }
}

/****************************************/
/*         Footbot Node Motion          */
/****************************************/
void CFootBotPhysarum::Node_Motion(){
   if (m_robot_type != STATE_NODE)
   {
      m_robot_type = STATE_NODE;
      m_pcRABA->SetData(0, m_robot_type);
   }
   if(e_f == 1)
   {
      m_pcLEDs->SetAllColors(CColor::BLUE);
   }
   else
   {
      m_pcLEDs->SetAllColors(CColor::WHITE);
   }
   Caculate_C_Node();
   Check_Chain();

   Update_Root_ID();
   Caculate_Envir_Factor();
   Spread_e_f();
}

/*Caculate the C_node when the value of it is null --255*/
void CFootBotPhysarum::Caculate_C_Node()
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   if(! tMsgs.empty())
   {
      for(size_t i=0; i < tMsgs.size();i++)
      {
         if(tMsgs[i].Data[0] == STATE_NODE || tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
         {
            // If there is a node which gradient value is just worth 1 more than mine
            if(tMsgs[i].Data[1] == m_node_gradient + 1)
            {
               C_node = tMsgs[i].Data[2];
            }
         }
      }
   }
}

void CFootBotPhysarum::Check_Chain()
{
   // Determine whether the P_node exist
   bool p_exist_flag = Lost_P_node();
   // Determine whether the P_node exist
   bool c_exist_flag = Lost_C_node();

   if(c_exist_flag)  C_node = 255;
   // Updating the gradient when it is null or the P_node is lost
   if(p_exist_flag || m_node_gradient == 255)
   {
      Update_gradient();
   }   
}

bool CFootBotPhysarum::exist_explorer()
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   bool flag = false;
   for(size_t i = 0; i < tMsgs.size();i++)
   {
      if(tMsgs[i].Data[0] == STATE_EXPLORER)
      {
         flag = true;      // exist explorer
         // break;
      }
   }
   return flag;
}

/*Determine whether the current node is the end of the robot chain and then caculate the environmental assessment factors*/
void CFootBotPhysarum::Update_Root_ID()
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();

   check_num_c_node();
   // if there are not only one child node and the node is not the end of the robot chain
   if(num_C_node > 1)
   {
      UInt8 change_m_e_f = 1;    // Determine whether all child nodes are stay in -1
      for(size_t i=0; i<tMsgs.size(); i++)
      {
         if(tMsgs[i].Data[0] == STATE_NODE || tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
         {
            if((tMsgs[i].Data[1] == m_node_gradient + 1) && (tMsgs[i].Data[5] != 0))
            {
               change_m_e_f = 0;
               break;
            }
         }
      }
      // Update the Data[5] for the node which has lots of child node
      // all the chil nodes are stay in -1 and there are not exist explorer
      bool flag = exist_explorer();
      if(change_m_e_f == 1 && flag == false)
      {
         e_f = 0;
         m_pcRABA->SetData(5,e_f);
      }
      else
      {
         e_f = 1;
         m_pcRABA->SetData(5,e_f);
      }
   }
   else     // there are only have one child node
   {
      for(size_t i=0; i<tMsgs.size(); i++)
      {
         if(tMsgs[i].Data[0] == STATE_NODE || tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
         {
            if(tMsgs[i].Data[2] == P_node )
            {
               if(tMsgs[i].Data[3] > 1)
               {
                  m_root_id = m_id;
                  m_pcRABA->SetData(4,m_root_id);
               }
               else
               {
                  m_root_id = tMsgs[i].Data[4];
                  m_pcRABA->SetData(4,m_root_id);
               }
            }
         }
      }
   }
}

/*Determine whether the number of C_node is change*/
void CFootBotPhysarum::check_num_c_node()
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   UInt8 flag = 0;
   UInt8 cur_sum_c_node = 0;  // Total number of current child nodes
   // bool change = false;      // 1 lost the P_node;   2 lost the C_node;
   for(size_t i = 0; i < tMsgs.size(); i++)
   {
      if(tMsgs[i].Data[0] == STATE_NODE ||  tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
      {
         if(tMsgs[i].Data[1] == m_node_gradient + 1)
         {
            cur_sum_c_node++;
         }
      }
   }
   // if there was no any child node exist
   if(num_C_node == 0)
   {
      num_C_node = cur_sum_c_node;
   }
   else
   {
      // the number of current child nodes is different from the previous number
      if(cur_sum_c_node != num_C_node)
      {
         sum_c_counter++;
      }
      else
      {
         sum_c_counter = 0;
      }
   }
   if(sum_c_counter > 300)
   {
      num_C_node = cur_sum_c_node;     // Update the value
      m_pcRABA->SetData(3, num_C_node);
      sum_c_counter = 0;
      // change = true;
   }
   // return change;
}

/*Update the root id*/
// void CFootBotPhysarum::Update_Root_ID()
// {
//    /* Get RAB messages from nearby foot-bots */
//    const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
//    for(size_t i = 0; i < tMsgs.size();i++)
//    {
//       if(tMsgs[i].Data[0] == STATE_NODE || tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
//       {
//          // Caculate the root_id
//          if (tMsgs[i].Data[2] == P_node)
//          {
//             m_root_id = tMsgs[i].Data[4];
//             if(tMsgs[i].Data[3] > 1)
//             {
//                m_root_id = m_id;       // update the root id
//             }
//          }
//       }
//    }
//    m_pcRABA->SetData(4,m_root_id);
// }

/*Determine whether the current node is the end of the robot chain and then caculate the environmental assessment factors*/
void CFootBotPhysarum::Caculate_Envir_Factor()
{
   UInt8 flag_exist_explorer = 0;
   if (C_node == 255)               // The end of the robot chain
   {
      bool flag = exist_explorer();

      if(flag)
      {
         counter_envir += 1;
      }
      else
      {
         counter_envir = 0;
      }
      if(counter_envir > 500)
      {
         e_f = 0;
         m_pcRABA->SetData(5,e_f);
         counter_envir = 0;
      }
   }
}

void CFootBotPhysarum::Spread_e_f()
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   if(C_node != 255 && num_C_node == 1)
   {
      for(size_t i=0;i<tMsgs.size();i++)
      {
         if(tMsgs[i].Data[4] == m_root_id && (tMsgs[i].Data[0] == STATE_NODE || tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY))
         {
            if(tMsgs[i].Data[2] ==  C_node)
            { 
               e_f = tMsgs[i].Data[5];
            }
         }
      }
      m_pcRABA->SetData(5,e_f);
   }
}

/****************************************/
/*    Footbot Temporary Node Motion     */
/****************************************/
void CFootBotPhysarum::Temporary_Motion()
{
   if (m_robot_type != STATE_TEMPORARY)
   {
      m_robot_type = STATE_TEMPORARY;
      m_pcRABA->SetData(0, m_robot_type);
   }

   m_pcLEDs->SetAllColors(CColor::YELLOW);

   m_pcWheels->SetLinearVelocity(0,0); //Stop

   counter += 1;        // Timer plus 1

   UInt8 num_node = 0;  // Init the number of node nearby

   Real dist_to_closest_node = 255;

   UInt8 flag_out = 1;
   UInt8 flag_exist_p_node = 0;

   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();

   for(size_t i = 0; i < tMsgs.size(); i++)
   {
      if(tMsgs[i].Data[0] == 1)
      {
         // There is a target
         flag_target = 1;
      }
      else if(tMsgs[i].Data[0] == STATE_NODE || tMsgs[i].Data[0] == STATE_TEMPORARY || tMsgs[i].Data[0] == 0)
      {
         if(tMsgs[i].Data[5] == 0)
         {
            flag_out = 0;
            break;
         }
         if(tMsgs[i].Data[2] == P_node)
         {
            flag_exist_p_node = 1;
         }   
         if(tMsgs[i].Range < dist_to_closest_node)
         {
            dist_to_closest_node = tMsgs[i].Range;
         }
      }
   }

   if(flag_out == 0)
   {
      counter = 0;
      Node_to_Other_State();
      Explorer_Motion();
   }
   if(m_id == 37) std::cout << "counter:" << counter << "\n";
   // -- If 10 seconds have expired, make a decision
   if(counter > 100)
   {
      counter = 0;
      if (flag_target == 1)
      {
         // Node_to_Other_State();
         m_node_gradient = 255;
         Node_Motion();
      }
      else if(flag_target == 0 && dist_to_closest_node > 0.6*RAB_RANGE)
      {
         // Node_to_Other_State();
         m_node_gradient = 255;
         Node_Motion();
      }
      else
      {
         // if(m_id == 29) std::cout << "dist_to_closest:" << dist_to_closest_node << "\n";
         Node_to_Other_State();
         Explorer_Motion();
      }
   }
}

/*Whether the robot */
UInt8 CFootBotPhysarum::determine_chain_state()
{
   UInt8 flag = 0;      // 0 -- all the node are stay in the same value of the Data[5];       1  -- nodes are stay in different Data[5]
   UInt8 flag_a = 0;    // determine whether there was data[5] = 1
   UInt8 flag_b = 0;    // determine whether there was data[5] = 0
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   for(size_t i = 0; i < tMsgs.size(); ++i)
   {
      if(tMsgs[i].Data[0] == STATE_NODE ||  tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
      {
         if(tMsgs[i].Data[5] == 1)
         {
            flag_a = 1;
         }
         else if(tMsgs[i].Data[5] == 0)
         {
            flag_b = 1;
         }
      }
   }
   if(flag_a == 1 && flag_b == 1)
   {
      flag = 1;
   }
   return flag;
}

/*
 * This function is a generalization of the potential
 */
CVector2 CFootBotPhysarum::GeneralizedPotential(Real aver_gradient, Real sum_size)
{
   Real field_strength;
   CVector2 potential_direction = {0,0};
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   UInt8 flag =  determine_chain_state();
   UInt8 temp_max_gradient = 0;  // store the max gradient when the Data[5] is 1
   UInt8 index = 255;
   for(size_t i = 0; i < tMsgs.size(); ++i) 
   {
      if(tMsgs[i].Data[0] == STATE_NODE ||  tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
      {
         if(flag == 0)
         {
            if(tMsgs[i].Data[5] == 1)
            {
               field_strength = (tMsgs[i].Data[1] - aver_gradient)*log(tMsgs[i].Range)/tMsgs[i].Data[5];
            }
            else
            {
               field_strength = -(tMsgs[i].Data[1] - aver_gradient)*log(tMsgs[i].Range);
            }
            potential_direction += CVector2(field_strength,tMsgs[i].HorizontalBearing);
         }
         else      // nodes are stay in different Data[5]
         {
            if(tMsgs[i].Data[5] == 1 && tMsgs[i].Data[1] > temp_max_gradient)
            {
               temp_max_gradient = tMsgs[i].Data[1];
               index = i;
            }
         }
      }
   }
   if (flag == 0)
   {
      potential_direction /=sum_size;
   }
   else
   {
      potential_direction = CVector2(1,tMsgs[index].HorizontalBearing);
   }
   return potential_direction;
}

/*Caculate the balance factor
*/
Real CFootBotPhysarum::Get_Balance_Factor()
{
   /* Get readings from proximity sensor */
   const CCI_FootBotProximitySensor::TReadings& tProxReads = m_pcProximity->GetReadings();

   Real max_Prox_Value = 0;

   for(size_t i=0; i< tProxReads.size();i++)
   {
      if(tProxReads[i].Value > max_Prox_Value)
      {
         max_Prox_Value = tProxReads[i].Value;
      }
   }
   return max_Prox_Value;
}

/*
 * This function is a generalization of the advoid obstacles based on Guassian
 */
 CVector2 CFootBotPhysarum::Generalized_Dynamic_Advoid_Obstacles()
 {
   /* Get readings from proximity sensor */
   const CCI_FootBotProximitySensor::TReadings& tProxReads = m_pcProximity->GetReadings();
   /*Store the sensing value of proximity sensor*/
   Real fitness_prox[tProxReads.size()]={0};  // Init the value of array = 0

   UInt8 flag_obstacles = 0;     // 0 means no obstacle detected; 1 indicates an obstacle has been detected

   CVector2 advoid_direction;

   for(size_t i=0;i < tProxReads.size();++i)
   {
      /*Calculate the effect of other proximity sensors*/
      if(!m_cGoStraightAngleRange.WithinMinBoundIncludedMaxBoundIncluded(tProxReads[i].Angle) && tProxReads[i].Value > m_fDelta )
      {
         flag_obstacles = 1;
         for(size_t j=0; j < tProxReads.size();j++)
         {
            // calculate the effect of other proximity sensors based on the Gaussian function
            Real g_result = Generalized_Guassian(tProxReads[j].Angle.GetValue(),tProxReads[i].Angle.GetValue(),1);
            fitness_prox[j] += tProxReads[i].Value * g_result;
         }
      }
   }

   /* find the minimum value of guassion function and the index of proximity*/
   Real min_g_result = 255;
   size_t min_g_result_index = 0;
   if(flag_obstacles != 0)
   {
      for(size_t i = 0; i < tProxReads.size(); i++)
      {
         if (fitness_prox[i] < min_g_result)
         {
            min_g_result = fitness_prox[i];
            min_g_result_index = i;
         }
         // std::cout<< "pro:" << i << "a_s:" << fitness_prox[i] << "prox_value:" << tProxReads[i].Value << "\n";
      }
      advoid_direction = CVector2(1,tProxReads[min_g_result_index].Angle);
   }
   else
   {
      advoid_direction = {0,0};
   }
   return advoid_direction;
 }

CVector2 CFootBotPhysarum::repulsion_vector()
{
   /* Get readings from proximity sensor */
   const CCI_FootBotProximitySensor::TReadings& tProxReads = m_pcProximity->GetReadings();

   /* Sum them together */
   /* length*cos(angle),length*sin(angle)*/
   CVector2 cAccumulator;
   for(size_t i = 0; i < tProxReads.size(); ++i) {
      cAccumulator += CVector2(tProxReads[i].Value, tProxReads[i].Angle);
   }
   cAccumulator = -cAccumulator/tProxReads.size();     // tProxReads.size() = 24
   /* If the angle of the vector is small enough and the closest obstacle
    * is far enough, continue going straight, otherwise curve a little
    */
   
   // CRadians cAngle = cAccumulator.Angle();
   // if(m_cGoStraightAngleRange.WithinMinBoundIncludedMaxBoundIncluded(cAngle) && cAccumulator.Length() < m_fDelta) {
   //    cAccumulator = {0,0};
   // }
   // std::cout << "vector:" << cAccumulator << "\n";
   return cAccumulator;
}

/*
 * This function is a generalization of the Guassian function
 * x -- the input value, m -- the mean value,   s --variance
 */
 Real CFootBotPhysarum::Generalized_Guassian(Real x, Real m, Real s)
 {
    static const Real inv_sqrt_2pi = 0.3989422804014327;
    Real a = (x - m) / s;

    return inv_sqrt_2pi / s * std::exp(-0.5f * a * a);
}

/* Transforms the given direction vector into wheel actuation
  This version does not consider the length of the direction vector*/
void CFootBotPhysarum::Vector_to_Wheel_Velocity_Noscale(CVector2 direction)
{
    // Get the direction angle in (-pi,pi)
   // Real angle = direction.Angle();
   Real angle = atan2(direction.GetY(),direction.GetX());
    // Declare speeds
   Real lvel, rvel;
   if (angle > 0)
   {
      lvel = m_fWheelVelocity * (1 - 2 * angle  / M_PI);
      rvel = m_fWheelVelocity;
   }
   else
   {
      lvel = m_fWheelVelocity;
      rvel = m_fWheelVelocity * (1 + 2 * angle  / M_PI);
   }
    // Set wheel speeds
    m_pcWheels->SetLinearVelocity(lvel, rvel);
   //  m_pcWheels->SetLinearVelocity(m_fWheelVelocity
}

/* Transforms the given direction vector into wheel actuation
	  This version scales the speed with the length of the direction vector*/
// void CFootBotPhysarum::Vector_to_Wheel_Velocity_scale(CVector2 direction)
// {
//     // Get the direction angle in (-pi,pi)
//    // Real angle = direction.Angle();
//    Real angle = atan2(direction.GetY(),direction.GetX());

//    // Get the length and scale it with 10 (the direction vector is often very short)
// 	Real length = 10 * math.sqrt(direction.GetX()*direction.GetX() + direction.GetY()*direction.GetY())
//     // Declare speeds
//    Real lvel, rvel;
//    if (angle > 0)
//    {
//       lvel = m_fWheelVelocity * (1 - 2 * angle  / M_PI);
//       rvel = m_fWheelVelocity;
//    }
//    else
//    {
//       lvel = m_fWheelVelocity;
//       rvel = m_fWheelVelocity * (1 + 2 * angle  / M_PI);
//    }
//     // Set wheel speeds
//     m_pcWheels->SetLinearVelocity(lvel, rvel);
// }

/*Update the Gradient of node*/
void CFootBotPhysarum::Update_gradient()
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   UInt8 min_gradient = 255;
   UInt8 min_g_id = 255;
   if(! tMsgs.empty()) {
      for(size_t i = 0; i < tMsgs.size(); i++)
      {
         if(tMsgs[i].Data[0] == STATE_NODE ||  tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
         {
            if(C_node != 255)
            {
               if(tMsgs[i].Data[1] < min_gradient && tMsgs[i].Data[2] != C_node)
               {
                  min_gradient = tMsgs[i].Data[1];
                  min_g_id     = tMsgs[i].Data[2];
               }
            }
            else
            {
               if(tMsgs[i].Data[1] < min_gradient)
               {
                  min_gradient = tMsgs[i].Data[1];
                  min_g_id     = tMsgs[i].Data[2];
               }
            }
         }
      }
      
      if (m_node_gradient != min_gradient + 1 && min_gradient != 255)
      {
         // std::cout << "id:" << m_id << "g:" << m_node_gradient << "update_g:" << min_gradient + 1 << "\n";
         m_node_gradient = min_gradient + 1;
         m_pcRABA->SetData(1, m_node_gradient);
         if(m_robot_type != STATE_TEMPORARY)         P_node = min_g_id;
         // std::cout << "gadient:" << m_node_gradient << "\n";
      }
      else if(min_g_id == 255)      // Could not find the new P_node
      {  
         Node_to_Other_State();
         // std::cout << "ID:" << m_id << "p_n:" << P_node << "????????" <<"\n";
         Lost_Chain_Motion();
      }
   }
}

bool CFootBotPhysarum::Lost_P_node()
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   UInt8 flag = 0;      // the P_node exist;    2 the C_node exist;
   bool lost = false;      // 1 lost the P_node;   2 lost the C_node;
   for(size_t i = 0; i < tMsgs.size(); i++)
   {
      if(tMsgs[i].Data[0] == STATE_NODE ||  tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
      {
         if(tMsgs[i].Data[2] == P_node)
         {
            flag = 1;
            break;
         }
      }
   }
   if(flag == 1)
   {
      p_counter = 0;
   }
   else
   {
      p_counter++;
   }
   if(p_counter > 300)
   {
      p_counter = 0;
      lost = true;
   }
   return lost;
}

bool CFootBotPhysarum::Lost_C_node()
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   UInt8 flag = 0;      // the P_node exist;    2 the C_node exist;
   bool lost = false;      // 1 lost the P_node;   2 lost the C_node;
   if(C_node != 255){
      for(size_t i = 0; i < tMsgs.size(); i++)
      {
         if(tMsgs[i].Data[0] == STATE_NODE ||  tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_TEMPORARY)
         {
            if(tMsgs[i].Data[2] == C_node)
            {
               flag = 1;
               break;
            }
         }
      }
      if(flag == 1)
      {
         c_counter = 0;
      }
      else
      {
         c_counter++;
      }
      if(c_counter > 300)
      {
         c_counter = 0;
         lost = true;
      }
      return lost;
   }
}

void CFootBotPhysarum::Node_to_Other_State()
{
   m_node_gradient = 255;
   num_C_node = 0;
   m_root_id = 255;
   P_node = 255;
   C_node = 255;
   e_f = 1;

   m_pcRABA->SetData(1, m_node_gradient);
   m_pcRABA->SetData(3, num_C_node);
   m_pcRABA->SetData(4,m_root_id);
   m_pcRABA->SetData(5,e_f);
}

/****************************************/
/*          Footbot Lost Chain          */
/****************************************/
void CFootBotPhysarum::Lost_Chain_Motion()
{
   if (m_robot_type != STATE_LOST_CHAIN)
   {
      m_robot_type = STATE_LOST_CHAIN;
      m_pcRABA->SetData(0, m_robot_type);
   }
   m_pcLEDs->SetAllColors(CColor::GREEN);
   // m_pcWheels->SetLinearVelocity(0,0);

   /*Calculate the advoid obstacles direction*/
   // CVector2 advoid_direction = Generalized_Dynamic_Advoid_Obstacles();
   CVector2 advoid_direction = repulsion_vector();
   /*Moving according to the direction*/
   // Vector_to_Wheel_Velocity_Noscale(advoid_direction);

   Lost_to_Explorer();

   Real temp = advoid_direction.GetX()*advoid_direction.GetX()+advoid_direction.GetY()*advoid_direction.GetY();
   if (temp > 0.001)
   {
      Vector_to_Wheel_Velocity_Noscale(advoid_direction);
   }
   else
   {
      m_pcWheels->SetLinearVelocity(m_fWheelVelocity,m_fWheelVelocity);
   }
   
}

/*The condition of */
void CFootBotPhysarum::Lost_to_Explorer()
{
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();

   UInt8 flag_chain = 0;

   for(size_t i = 0; i < tMsgs.size(); i++)
   {
      if(tMsgs[i].Data[0] == STATE_NODE || tMsgs[i].Data[0] == STATE_TEMPORARY || tMsgs[i].Data[0] == 0)
      {
         flag_chain = 1;
         break;
      }
   }
   if(flag_chain == 1)
   {
      Explorer_Motion();
   }
}

/****************************************/
/*       Footbot advoid obstacles       */
/****************************************/
void CFootBotPhysarum::Advoid_obstacles() {
   /* Get readings from proximity sensor */
   const CCI_FootBotProximitySensor::TReadings& tProxReads = m_pcProximity->GetReadings();
   
   /* Sum them together */
   /* length*cos(angle),length*sin(angle)*/
   CVector2 cAccumulator;
   for(size_t i = 0; i < tProxReads.size(); ++i) {
      cAccumulator += CVector2(tProxReads[i].Value, tProxReads[i].Angle);
   }
   cAccumulator /= tProxReads.size();     // tProxReads.size() = 24
   /* If the angle of the vector is small enough and the closest obstacle
    * is far enough, continue going straight, otherwise curve a little
    */
   CRadians cAngle = cAccumulator.Angle();
   if(m_cGoStraightAngleRange.WithinMinBoundIncludedMaxBoundIncluded(cAngle) &&
      cAccumulator.Length() < m_fDelta ) {
      /* Go straight */
      m_pcWheels->SetLinearVelocity(m_fWheelVelocity, m_fWheelVelocity);
   }
   else {
      /* Turn, depending on the sign of the angle */
      if(cAngle.GetValue() > 0.0f) {
         m_pcWheels->SetLinearVelocity(m_fWheelVelocity, 0.0f);
      }
      else {
         m_pcWheels->SetLinearVelocity(0.0f, m_fWheelVelocity);
      }
   }
}

void CFootBotPhysarum::Reset()
{
   Init_RABA_Data();
}

void CFootBotPhysarum::Init_RABA_Data()
{
   m_pcRABA->ClearData();
   /* Switch to state explorer */
   m_robot_type = STATE_WANDER;
   /* Tell robots around that this robot is explorer */
   m_pcRABA->SetData(0, m_robot_type);

   const std::string& str = CCI_Controller::GetId();
   // std::cout<<str.find("r")<<"\n";
   std::string str3 = str.substr(8,str.length());
   // std::cout << str <<" "<< str3 << "\n";

   m_id = atoi(str3.c_str());

   m_pcRABA->SetData(2,m_id);

   m_pcRABA->SetData(5,e_f);

   // std::cout << "data7:" << 
}

// --[[ Logic of state WANDER IN NEST ]]
void CFootBotPhysarum::wander_in_waiting_area()
{
   if (m_robot_type != STATE_WANDER)
   {
      m_robot_type = STATE_WANDER;
      m_pcRABA->SetData(0, STATE_WANDER);
   }

   if(should_exit())
   {
      // The robot should exit the nest because there is
	  	// a nest nearby, or because other robots are
	  	// already exiting
      
      Exist_Motion();
   }
   if (is_out_waiting_area())
   {
      // The robot just exited the waitiing area
		// It's the first, so become a explorer
      m_pcRABA->SetData(3,0);
      m_pcRABA->SetData(4,0);
		Explorer_Motion();
   }
   CVector2 advoid_direction = repulsion_vector();

   /*Moving according to the direction*/

   Real temp = advoid_direction.GetX()*advoid_direction.GetX()+advoid_direction.GetY()*advoid_direction.GetY();
   if ( temp > 0.001)
   {
      Vector_to_Wheel_Velocity_Noscale(advoid_direction);
   }
   else
   {
      m_pcWheels->SetLinearVelocity(5,5);
   }
}

// [[ Returns true if there is nest in range, or if other robots are in state EXIT_NEST ]]
bool CFootBotPhysarum::should_exit()
{
   bool flag = false;
   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   if(! tMsgs.empty())
   {
      for(size_t i = 0; i < tMsgs.size();i++)
      {
         if(tMsgs[i].Data[0] == 0 || tMsgs[i].Data[0] == STATE_EXIST)
         {
            flag = true;
         }
      }
   }
   else  // No messages received
   {
      flag = false;
   }
   return flag;
}

void CFootBotPhysarum::Exist_Motion()
{
   if (m_robot_type != STATE_EXIST)
   {
      m_robot_type = STATE_EXIST;
      m_pcRABA->SetData(0, STATE_EXIST);
   }

   UInt8 dist = 255;
   CRadians angle;
   bool direction = false;
   UInt8 sum_exist = 0;
   CVector2 accum = {0,0};

   /* Get RAB messages from nearby foot-bots */
   const CCI_RangeAndBearingSensor::TReadings& tMsgs = m_pcRABS->GetReadings();
   if(!tMsgs.empty())
   {
      for(size_t i = 0; i < tMsgs.size();i++)
      {
         if((tMsgs[i].Data[0] == STATE_NODE || tMsgs[i].Data[0] == STATE_TEMPORARY || tMsgs[i].Data[0] == 0) && tMsgs[i].Range < dist)
         {
            dist = tMsgs[i].Range;
            angle = tMsgs[i].HorizontalBearing;
            direction = true;
         }
         else if(tMsgs[i].Data[0] == STATE_EXIST)
         {
            // Exiting robot detected
            sum_exist++;
            // Calculate distance to nest of this robot
            Real nest_dist = tMsgs[i].Data[4] + tMsgs[i].Data[3] * 255 + tMsgs[i].Range;
            if (nest_dist < dist)
            {
               // Found the robot who knows the closest way to a nest
               dist = nest_dist;
               angle = tMsgs[i].HorizontalBearing;
               direction = false;
            }
            // Calculate the contribution of this robot
			   Real lj = lennard_jones(tMsgs[i].Range,EXITING_DISTANCE,EXITING_GAIN);
            
            CVector2 contr = CVector2(lj,tMsgs[i].HorizontalBearing);

            accum = accum + contr;
         }
      }
      // State transition logic
	   // If you can see the nest directly and you're out of the nest, explore
      if(direction && is_out_waiting_area())
      {
         m_pcRABA->SetData(3,0);
         m_pcRABA->SetData(4,0);
         Explorer_Motion();
         return;
      }
      // State logic
	   // Postprocess the data collected
	   // Take the average of the exiting robot interaction
      if(sum_exist > 1) 
      {
         accum = accum/sum_exist;
      }
      // Calculate the LJ interaction to the nest
	   CVector2 nest_contr = { 0,0 };
	   if (dist < 255)
      {
         UInt8 magnitude = 2;
         if(dist < 1.5*RAB_RANGE)
         {
            magnitude = lennard_jones(dist,NEST_DISTANCE,NEST_GAIN);
         }
         nest_contr = CVector2(magnitude,angle);
         // send around the closest direction to nest known
         m_pcRABA->SetData(3,dist/255);
         m_pcRABA->SetData(4,dist%255);
      }
      else
      {
         m_pcRABA->SetData(3,255);
         m_pcRABA->SetData(4,255);
      }
		// Calculate the direction
      CVector2 sum_direction = accum+nest_contr;
   	// Actuate wheels
	   Vector_to_Wheel_Velocity_Noscale(sum_direction);
   }
}

bool CFootBotPhysarum::is_out_waiting_area()
{
   // The robot is out of the nest when all the motor ground
	// sensors see white (reading = 1)
   bool flag_out = true;
   const CCI_FootBotMotorGroundSensor::TReadings& tGroundReads = m_pcMGS->GetReadings();
   for(size_t i = 0; i < tGroundReads.size();i++)
   {
      if(tGroundReads[i].Value != 1)
      {
         flag_out = false;
         break;
      }
   }
   return flag_out;
}

/**[[ Calculates the Lennard-Jones force intensity, given the interaction distance
	  Instead of the classical exponents 12 and 6, here we use 4 and 2 to avoid
	  overflows. See: http://en.wikipedia.org/wiki/Lennard-Jones_potential	]]**/
Real CFootBotPhysarum::lennard_jones(Real dist, Real target_dist,Real gain)
{
	Real factor = target_dist / dist;
	Real factor2 = factor * factor;
	Real result = (gain / dist) * (factor2 - factor2 * factor2);
	// For values of dist close to zero, the function goes to -inf => clamp the value to prevent explosions
	if(result < -5)
   {
      return -5;
   }
   else 	return result;
}

/****************************************/
/****************************************/

/*
 * This statement notifies ARGoS of the existence of the controller.
 * It binds the class passed as first argument to the string passed as
 * second argument.
 * The string is then usable in the configuration file to refer to this
 * controller.
 * When ARGoS reads that string in the configuration file, it knows which
 * controller class to instantiate.
 * See also the configuration files for an example of how this is used.
 */
REGISTER_CONTROLLER(CFootBotPhysarum, "footbot_physarum_controller")
